#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <poll.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>

#include "picohttpparser.h"

#include "event_loop_internal.h"

void event_loop_conn_io_res(event_loop_conn *conn) {
  while (1) {
    ssize_t res = 0;
    size_t remain = conn->wbuf_size - conn->wbuf_sent;

    do {
      res = write(conn->fd, &conn->wbuf[conn->wbuf_sent], remain);
    } while (res < 0 && errno == EINTR);

    // EAGAIN doesn't mean there was an error, but rather that there's no more
    // data right now, but there might be more later, aka "try again later"
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    // If it's not EGAIN, there was an error writing so we simply end the
    // request
    if (res < 0) {
      conn->state = event_loop_conn_state_end;
      return;
    }

    conn->wbuf_sent += (size_t)res;

    // After writing a response, we switch back to request mode to process a
    // next request
    if (conn->wbuf_sent == conn->wbuf_size) {
      conn->state = conn->close_after_write ? event_loop_conn_state_end
                                            : event_loop_conn_state_req;
      /* c->wbuf_sent = 0; */
      /* c->wbuf_size = 0; */

      return;
    }
  }
}

bool event_loop_handle_request(event_loop_conn *conn) {
  // Prevents the request handler function from looping indefinitely without
  // ever consuming new data
  if (conn->rbuf_size == 0) {
    return false;
  }

  /* if (conn->process_func != NULL) { */
  /*   conn->process_func(conn); */
  /* } */

  const char *method, *path;
  struct phr_header headers[16];
  size_t method_len, path_len, num_headers;
  int minor_version;

  num_headers = sizeof(headers) / sizeof(headers[0]);

  int res = phr_parse_request((const char *)conn->rbuf, conn->rbuf_size,
                              &method, &method_len, &path, &path_len,
                              &minor_version, headers, &num_headers, 0);

  if (res > 0) {
    /* for (size_t i = 0; i < num_headers; i++) { */
    /*   /1* printf("%.*s: ", headers[i].name_len, headers[i].name); *1/ */
    /*   /1* printf("%.*s\n", headers[i].value_len, headers[i].value); *1/ */
    /*   if (strncmp("Connection", headers[i].name, headers[i].name_len) == 0) { */
    /*     if (strncmp("close", headers[i].value, headers[i].value_len) == 0) { */
    /*       conn->close_after_write = true; */
    /*       break; */
    /*     } */
    /*   } */
    /* } */

    /* memcpy(conn->wbuf, http_500, http_500_len); */

    /* // Move the new request up to the front of the read buffer */
    /* memmove(conn->rbuf, &conn->rbuf[res], conn->rbuf_size - res); */
    /* conn->rbuf_size -= res; */

    /* conn->state = event_loop_conn_state_res; */
    /* conn->wbuf_size = http_500_len; */
    /* conn->wbuf_sent = 0; */

    /* /1* event_loop_conn_io_res(conn); *1/ */

    /* // End the connection if the http request is invalid, or when the response */
    /* // is too large to fit into a single read buffer */
  } else if (res == -1 ||
             (res == -2 && conn->rbuf_size == EVENT_LOOP_BUFFER_SIZE)) {
    conn->state = event_loop_conn_state_end;
  }

  return conn->state == event_loop_conn_state_req;
}

/**
 * Read new data into the read buffer. This command performs at most one
 * successful read syscall.
 *
 * Returns whether the function should be retried immediately or not.
 */
void event_loop_conn_io_req(event_loop_conn *conn) {
  while (1) {
    ssize_t res;
    size_t cap = EVENT_LOOP_BUFFER_SIZE - conn->rbuf_size;

    // Try to read at most cap bytes from the file descriptor
    do {
      res = read(conn->fd, &conn->rbuf[conn->rbuf_size], cap);
    } while (res < 0 && errno == EINTR);

    // EGAIN means we try again later
    if (res < 0 && errno == EAGAIN) {
      return;
    }

    // Any other negative error message means the read errored out
    if (res < 0) {
      conn->state = event_loop_conn_state_end;

      return;
    }

    // Output of 0 and no error means we've reached the end of the input. We run
    // try_one_request one more time and close the connection if this doesn't
    // change the result.
    if (res == 0) {
      conn->state = event_loop_conn_state_end;

      return;
    }

    // We switch to processing mode if we've reached the end of the data stream,
    // or if the read buffer is filled
    /* if (res == 0 || c->rbuf_size == MAX_MSG_SIZE) { */
    /*     c->state = STATE_PROCESS; */
    /*     return false; */
    /* } */

    conn->rbuf_size += (size_t)res;

    /* assert(conn->rbuf_size <= sizeof(conn->rbuf)); */

    // Try to process requests one by one.
    // Try to process requests one by one.
    // Why is there a loop? Please read the explanation of "pipelining".
    while (event_loop_handle_request(conn)) {
    };

    // We can keep reading as long as we're in request mode
    if (conn->state != event_loop_conn_state_req) {
      return;
    };
  }
}

void event_loop_conn_io(event_loop_conn *conn) {
  switch (conn->state) {
  case event_loop_conn_state_req:
    event_loop_conn_io_req(conn);
    break;
  case event_loop_conn_state_res:
    event_loop_conn_io_res(conn);
    break;
  case event_loop_conn_state_end:
    printf("we shouldn't be here\n");
    break;
  }
}
